"""Meetings and ratings, time preferences of/for them."""

from meety import io
from meety.meetings import preferences


def reset_options():
    io.reset_options()
    preferences.reset_options()


def set_options(config, name):
    io.set_options(config, name)
    preferences.set_options(config.get("preferences") or {}, name)



import logging
import traceback

logging.basicConfig(format="%(levelname)s: %(message)s")


class MeetyLogger:
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG

    LEVEL_NAMES = {
        ERROR: "Error",
        WARNING: "Warning",
        INFO: "Info",
        DEBUG: "Debug",
    }

    def __init__(self):
        self._logger = logging.getLogger()
        self._contents = []

    def setDebug(self):
        self._logger.setLevel(logging.DEBUG)

    def setInfo(self):
        self._logger.setLevel(logging.INFO)

    def setWarn(self):
        self._logger.setLevel(logging.WARN)

    def setError(self):
        self._logger.setLevel(logging.ERROR)

    def debug(self, message, condition=True):
        if not condition:
            return
        self._logger.debug(message)
        self._add(logging.DEBUG, message)

    def info(self, message, condition=True):
        if not condition:
            return
        self._logger.info(message)
        self._add(logging.INFO, message)

    def warning(self, message, condition=True):
        if not condition:
            return
        self._logger.warning(message)
        self._add(logging.WARNING, message)

    def error(self, message, condition=True):
        if not condition:
            return
        self._logger.error(message)
        self._add(logging.ERROR, message)

    def exception(self, ex, warn=True):
        msg = f"Exception: {ex}"
        if warn:
            self._logger.warning(msg)
        else:
            self._logger.debug(msg)
        self._logger.debug(traceback.format_exc())

    def propagate(self, flag):
        self._logger.propagate = flag

    def expected(self, what, name, data, error=False):
        text = (
            f"Ignoring data '{data}' for '{name}' "
            f"where {what} was expected."
        )
        if error:
            self.error(text)
        else:
            self.warning(text)

    def print_stack(self):
        traceback.print_stack()

    @classmethod
    def level_name(cls, level):
        return cls.LEVEL_NAMES[level]

    def contents(self, level=logging.WARNING):
        return [(lev, msg) for (lev, msg) in self._contents if lev >= level]

    def _add(self, level, message):
        self._contents.append((level, message))


log = MeetyLogger()



import pytest
import testdata

from meety import meetings
from meety.config import PartialConfig
from meety.loader import Loader

test_config = testdata.load(["test_loader.yaml"])


@pytest.mark.parametrize(
    "filename, further_filenames, expected, config",
    [
        (
            c["file"],
            c.get("further-files", []),
            c["expected"],
            c.get("config", {})
        ) for c in test_config
    ]
)
def test_loader(filename, further_filenames, expected, config):
    """Compare cleaned dicts of loaded meetings with expected dicts."""
    test_config = PartialConfig("test config", None, None)
    test_config._data = config
    test_config.apply()
    meetings.reset_options()
    meetings.set_options(test_config.meetings, "test_config")

    loader = Loader()
    _load_file(loader, filename)
    for f in further_filenames:
        _load_file(loader, f)
    assert [m.data for m in loader.meetings] == expected


def _load_file(loader, filename):
    loader.load_from_file(testdata.testpath(["data", filename]))



from PyQt5.QtWidgets import (
    QComboBox,
    QPlainTextEdit,
    QVBoxLayout,
    QWidget,
)

from meety.logging import log


class TabLog(QWidget):
    LEVELS = [
        log.WARNING,
        log.INFO,
        log.DEBUG,
    ]

    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._create_log_level())
        layout.addWidget(self._create_log_list())
        super().setLayout(layout)
        self._update_log()

    def _create_log_level(self):
        self._log_level = QComboBox()
        self._log_level.addItems([
            "warnings and errors",
            "informational",
            "debug messages",
        ])
        return self._log_level

    def _create_log_list(self):
        self._log_list = QPlainTextEdit()
        self._log_list.setReadOnly(True)
        return self._log_list

    def _connect_widget_signals(self):
        self._log_level.currentIndexChanged.connect(self._update_log)

    def _update_log(self):
        level = self.LEVELS[self._log_level.currentIndex()]
        text = "\n".join([
            f"{log.level_name(lev)}: {msg}"
            for (lev, msg) in log.contents(level)
        ])
        self._log_list.setPlainText(text)



import subprocess

from meety.connect import handlers as connect_handlers
from meety.logging import log

handlers = connect_handlers.Handlers()


def set_options(config):
    """Set options for module `connect`."""
    if isinstance(config, dict):
        connect_handlers.set_options(
            handlers,
            config.get("handlers") or []
        )
    else:
        log.expected("dictionary", "connect", config)


def start(handler, test_run=False):
    """Start a meeting by running the provided command."""
    if test_run:
        return ""
    try:
        log.info(f"Run '{handler.cmd}' with options '{handler.options}'")
        subprocess.Popen(handler.cmd.split(" "), **handler.options)
        log.info("... successful.")
    except FileNotFoundError:
        log.info("... failed.")
        error = f"Failed to execute handler '{handler.cmd}'!"
        log.error(error)
        return error
    else:
        return ""


def applicable_handlers(meeting):
    """Return a list of all applicable handlers, ordered by preference."""
    return list(generate_applicable_handlers(meeting))


def generate_applicable_handlers(meeting):
    """Return a generator for all applicable handlers,
    ordered by preference.
    """
    for handler in handlers.ordered:
        cmd, options = try_handler(handler, meeting)
        if cmd:
            yield connect_handlers.NamedHandler(
                handler.name,
                cmd,
                options
            )


def try_handler(handler, meeting):
    """Try to apply a handler based on the meeting's data.
    Returns the resulting command on success and `None` otherwise.
    """
    log.debug(f"Try handler '{handler.name}'.")
    try:
        cmd = handler.cmd.format_map(meeting.data)
    except KeyError:
        log.debug(f"Handler '{handler.name}' misses information.")
        return (None, None)
    else:
        return (cmd, handler.options)



import fnmatch
from collections import namedtuple

from meety import (
    io,
    system,
)
from meety.logging import log

NamedHandler = namedtuple("NamedHandler", "name cmd options")


def set_options(handlers, actions):
    if not isinstance(actions, list):
        log.expected("list", "connection handlers", actions)
        return

    for entry in actions:
        if isinstance(entry, dict):
            handlers.apply_action(entry)
        else:
            log.expected(
                "dictionary",
                "connection handler",
                entry
            )


class Handlers:
    """An ordered list of handlers.

    Handlers have a name (for reference) and a command. They are
    ordered linearly by preference. The preference order can be
    changed by the user: handlers can be removed, prepended and
    appended. New handlers can be registered as well.

    Changing of the preference is independent from the registration
    of a command for a handler. The intention behind this is that
    handlers should ideally be provided by the application while
    their preferred order is more subjective.
    """

    def __init__(self):
        """At the beginning, no handler is registered."""
        self._command = {}
        self._order = []

    def register(self, entry):
        """Register handler with a name and a command in `entry`."""
        name = entry["name"]
        cmd = entry["command"]
        options = {}
        options["shell"] = io.str_is_true(entry.get("shell", "False"))
        req_system = entry.get("system")
        req_app = entry.get("installed")
        if system.satisfies(req_system, req_app):
            self._command[name] = (cmd, options)
            self.append(entry)

    @property
    def ordered(self):
        """Yield all handlers with a command, ordered by preference."""
        for name in self._order:
            cmd, options = self._command.get(name)
            if cmd:
                yield NamedHandler(name, cmd, options)

    def prepend(self, entry):
        """Add handler with name in `entry` as the first handler.
        If the handler already occurs in the preference order, then
        the latter occurrence is removed.
        """
        name = entry["name"]
        self.remove(entry, True)
        self._order.insert(0, name)

    def append(self, entry):
        """Add handler with name in `entry`as the last handler.
        If the handler already occurs in the preference order, then
        it is not added. If you want to move to the end, remove it
        first.
        """
        name = entry["name"]
        if name not in self._order:
            self._order.append(name)

    def remove(self, entry, silent=False):
        """Remove all handlers with names matching the name in `entry`.
        If `silent` is `True`, then it will be logged if there is
        handler with a matching name.
        """
        name = entry["name"]
        matches = fnmatch.filter(self._order, name)
        for name in matches:
            log.debug(f"Remove handler '{name}'.")
            self._order.remove(name)
        if not matches and not silent:
            log.debug(f"No handler matches pattern '{name}'.")

    def apply_action(self, entry):
        """Try to apply action and log errors."""
        call = self._determine_action(entry)
        try:
            call(entry)
        except KeyError as e:
            log.warning(
                f"Failed to apply action '{entry}',"
                f" missing attribute {e}."
            )
        except Exception as e:
            log.warning(f"Failed to apply action '{entry}'.")
            log.exception(e)

    def _determine_action(self, entry):
        """Call the action defined by 'action' attribute."""
        action = entry.get("action")
        if not action:
            log.expected("attribute 'action'", "action", entry)
            return

        call = {
            "register": self.register,
            "prepend": self.prepend,
            "append": self.append,
            "remove": self.remove,
        }.get(action)
        return call

    def __str__(self):
        """List all handler definitions and the order of handlers."""
        return "\n".join([
            self._info_definitions(),
            "",
            self._info_order(),
        ])

    def _info_definitions(self):
        """Return string that lists all defined handlers (name: command)."""
        defs = len(self._command)
        info = [f"Handler definitions ({defs}):"]
        for (name, (cmd, options)) in self._command.items():
            if options == {}:
                opt_info = "no options"
            else:
                opt_info = ", ".join([
                    f"{k}={v}"
                    for (k, v) in options.items()
                ])
            info.append(f"  - {name}: {cmd} ({opt_info})")
        return "\n".join(info)

    def _info_order(self):
        """Return string that lists all handlers according to their order."""
        ords = len(self._order)
        info = [f"Order definitions ({ords}):"]
        for name in self._order:
            info.append(f"  - {name}")
        return "\n".join(info)



from contextlib import suppress

from meety.logging import log


class WeekdayTranslator:
    _translations = {}
    _preferred_languages = []

    @classmethod
    def reset_options(cls):
        cls._translations = {}
        cls._preferred_languages = []

    @classmethod
    def set_options(cls, config, name):
        WeekdayTranslation(name, config)

    @classmethod
    def add_translation(cls, translation):
        cls._translations[translation.language] = translation

    @classmethod
    def prefer(cls, language):
        cls._preferred_languages.append(language)

    @classmethod
    def all_languages(cls):
        return cls._translations.keys()

    @classmethod
    def ordered_languages(cls):
        for lang in cls._preferred_languages:
            yield lang
        for lang in cls.all_languages():
            if lang not in cls._preferred_languages:
                yield lang

    @classmethod
    def ordered_translations(cls):
        for lang in cls.ordered_languages():
            yield cls._translations[lang]

    @classmethod
    def get_name_by_number(cls, number):
        for trans in cls.ordered_translations():
            with suppress(KeyError):
                return trans.get_name_by_number(number)
        return str(number)

    @classmethod
    def get_number_by_name(cls, name):
        if not name:
            raise KeyError(name)
        clean_name = name.casefold()
        for trans in cls.ordered_translations():
            with suppress(KeyError):
                return trans.get_number_by_name(clean_name)
        try:
            value = int(name)
            assert 1 <= value <= 7
            return value
        except (ValueError, AssertionError):
            raise KeyError(name)

    @classmethod
    def debug_info(cls):
        return "\n".join([
            "Preferred languages:",
            cls._preferred_languages_to_str(),
            "Translations:",
            cls._translations_to_str(),
        ])

    @classmethod
    def _preferred_languages_to_str(cls):
        return "\n".join([
            f"  - {lang}"
            for lang in cls._preferred_languages
        ])

    @classmethod
    def _translations_to_str(cls):
        return "\n".join([
            f"{lang}\n--------------------\n{str(trans)}"
            for lang, trans in cls._translations.items()
        ])


class WeekdayTranslation:
    """Names for weekday numbers.
    Multiple names can be associated with a single number (for example,
    'Monday' and 'Mon' can be associated with integer 1). Each name
    can be translated to the associated integer but only the first name
    is returned on a given integer.
    """

    @classmethod
    def _ensure_valid_config(cls, config):
        valid_config = {}
        for entry in config.items():
            valid_entry = cls._ensure_valid_entry(entry)
            if valid_entry:
                num, names = valid_entry
                valid_config[num] = names
        return valid_config

    @classmethod
    def _ensure_valid_entry(cls, entry):
        num, names = entry
        num = int(num)
        if isinstance(names, str):
            return (num, [names])
        elif isinstance(names, list):
            return (num, names)
        else:
            log.expected(
                "string or list of strings",
                "names by number",
                names
            )
            return None

    def __init__(self, language, config):
        self._language = language
        config = self._ensure_valid_config(config)
        self._init_name_by_number(config)
        self._init_number_by_name(config)
        WeekdayTranslator.add_translation(self)

    def _init_name_by_number(self, config):
        self._name_by_number = {}
        for num, names in config.items():
            self._name_by_number[num] = names[0]

    def _init_number_by_name(self, config):
        self._number_by_name = {}
        for num, names in config.items():
            for name in names:
                clean_name = name.casefold()
                self._number_by_name[clean_name] = num

    @property
    def language(self):
        return self._language

    def get_name_by_number(self, number):
        return self._name_by_number[number]

    def get_number_by_name(self, name):
        return self._number_by_name[name]

    def __str__(self):
        return "\n".join([
            "Name by number:",
            self.name_by_number_to_str(),
            "Number by name:",
            self.number_by_name_to_str(),
        ])

    def name_by_number_to_str(self):
        return "\n".join([
            f"  {num}: {name}"
            for num, name in self._name_by_number.items()
        ])

    def number_by_name_to_str(self):
        return "\n".join([
            f"  {name}: {num}"
            for name, num in self._number_by_name.items()
        ])



try:
    import importlib.resources as py_resources
except ImportError:
    import importlib_resources as py_resources


def get_icon_path(filename):
    from meety.static import icons as static_icons
    return get_static_path(static_icons, filename)


def get_config_path(filename):
    from meety.static import config as static_config
    return get_static_path(static_config, filename)


def get_spec_path(filename):
    from meety.static import spec as static_spec
    return get_static_path(static_spec, filename)


def get_desktop_path(filename):
    from meety.static import desktop as static_desktop
    return get_static_path(static_desktop, filename)


def get_static_path(module, filename):
    with py_resources.path(module, filename) as path:
        return str(path)


def path(package, resource):
    return py_resources.path(package, resource)


def open_text(package, resource):
    return py_resources.open_text(package, resource)



import os

DEFAULT_MEETING_SPEC = os.path.join(
    os.path.expanduser("~"),
    "meetings.yaml"
)


def _add_argparser_common_arguments(parser, summary):
    parser.add_argument(
        "-f", "--file",
        dest="files",
        metavar="PATH",
        help="include meeting file",
        default=[],
        action="append",
    )
    parser.add_argument(
        "-d", "--directory",
        dest="directories",
        metavar="PATH",
        help="include YAML files in directory",
        default=[],
        action="append",
    )
    parser.add_argument(
        "-e", "--only-explicit",
        help="only include meeting files in the arguments",
        action="store_true",
    )
    parser.add_argument(
        "-p", "--copy-password",
        help="copy password to clipboard on connection",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--init",
        dest="init",
        metavar="PATH",
        help="create template meeting specification",
        nargs="?",
        const=DEFAULT_MEETING_SPEC,
        type=str,
    )
    parser.add_argument(
        "--verbose",
        help="increase logging verbosity",
        action="store_true",
    )
    parser.add_argument(
        "--debug",
        help="increase logging verbosity even more",
        action="store_true",
    )
    parser.add_argument(
        "-v", "--version",
        action="version",
        version=summary,
    )



from meety.io.actions import registry
from meety.io.actions.infer import InferAction
from meety.io.actions.replace import ReplaceAction
from meety.io.actions.strip import StripAction
from meety.io.actions.synonym import SynonymAction

InferAction.register()
StripAction.register()
SynonymAction.register()
ReplaceAction.register()



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QMenu

from meety import connect
from meety.io import copy_to_clipboard


class MeetingItemMenu(QMenu):
    handler_chosen = pyqtSignal(object, name="handler_chosen")
    reload_requested = pyqtSignal(name="reload_requested")

    def __init__(self, meeting):
        super().__init__()
        self._meeting = meeting
        self._counter = 0
        self._add_actions()

    def _add_actions(self):
        self._actions = {}
        if self._meeting:
            self.addAction("Copy &input data", self._on_copy_input)
            self.addAction("Copy &derived data", self._on_copy_derived)
            self._add_handler_actions()
            self.addSeparator()
        self.addAction("&Reload all meetings", self._on_reload)

    def _add_handler_actions(self):
        if self._meeting is None:
            return
        handlers = connect.applicable_handlers(self._meeting)
        if len(handlers) > 0:
            self.addSeparator()
        for handler in handlers:
            self._add_handler_action(handler)

    def _add_handler_action(self, handler):
        self._counter += 1
        number = str(self._counter)
        if self._counter < 10:
            # add shortcut for first entries,
            # distinguishable by the first digit
            number = "&" + number
        self.addAction(
            f"{number}. Connect via {handler.name}",
            lambda: self._on_handler(handler)
        )

    def _on_copy_input(self):
        text = self._meeting.input_data_to_str()
        copy_to_clipboard(text)

    def _on_copy_derived(self):
        text = self._meeting.data_to_str()
        copy_to_clipboard(text)

    def _on_handler(self, handler):
        self.handler_chosen.emit((handler, self._meeting))

    def _on_reload(self):
        self.reload_requested.emit()



import datetime
from contextlib import suppress

from meety.io import utils
from meety.io.datetime.weekday_translations import WeekdayTranslator
from meety.logging import log

_date_formats = []
_time_formats = []
_skip_words = []
_range_words = []
_list_words = []


def reset_options():
    _reset_options_date()
    _reset_options_time()
    _reset_options_skip_words()
    _reset_options_range_words()
    _reset_options_list_words()


def set_options(config, name):
    _set_options_date(config.get("date") or {})
    _set_options_time(config.get("time") or {})
    _set_options_weekday(config.get("weekday") or {}, name)
    _set_options_skip_words(config.get("skip words") or [])
    _set_options_range_words(config.get("range words") or [])
    _set_options_list_words(config.get("list words") or [])


def _reset_options_skip_words():
    global _skip_words
    _skip_words = []


def _set_options_skip_words(words):
    if isinstance(words, str):
        clean_words = [words]
    else:
        clean_words = words
    if not isinstance(clean_words, list):
        log.expected("string or list of strings", "skip words", words)
    for w in clean_words:
        if isinstance(w, str):
            _skip_words.append(f" {w.strip()} ")


def _reset_options_range_words():
    global _range_words
    _range_words = []


def _set_options_range_words(words):
    if isinstance(words, str):
        clean_words = [words]
    else:
        clean_words = words
    if not isinstance(clean_words, list):
        log.expected("string or list of strings", "range words", words)
    for w in clean_words:
        if isinstance(w, str):
            _range_words.append(f" {w.strip()} ")


def _reset_options_list_words():
    global _list_words
    _list_words = []


def _set_options_list_words(words):
    if isinstance(words, str):
        clean_words = [words]
    else:
        clean_words = words
    if not isinstance(clean_words, list):
        log.expected("string or list of strings", "list words", words)
    for w in clean_words:
        if isinstance(w, str):
            _list_words.append(f" {w.strip()} ")


def prepare_list_of_ranges(text):
    text = f" {text} "
    # prepend and append one space
    # to allow for matching words
    # at the beginning or the end

    text = utils.replace_words_by(text, _skip_words, " ")
    text = utils.replace_words_by(text, _range_words, " - ")
    text = utils.replace_words_by(text, _list_words, ", ")
    return text.strip()


def parse_timedelta(config):
    args = {k: int(v) for (k, v) in config.items()}
    try:
        delta = datetime.timedelta(**args)
    except TypeError as e:
        log.expected("valid arguments", "time delta", args)
        log.exception(e)
        return datetime.timedelta()
    else:
        return delta


def _reset_options_date():
    global _date_formats
    _date_formats = [
        "%Y-%m-%d",
        "%d.%m.%Y",
        "%d.%m.%y",
        "%d.%m.'%y",
    ]


def _set_options_date(config):
    formats = utils.expect_list_of_str(config.get("format"))
    for f in formats:
        _date_formats.append(f)


def parse_date(text, warn=True):
    clean_text = text.strip().casefold()
    for f in _date_formats:
        with suppress(ValueError):
            return datetime.datetime.strptime(clean_text, f)
    log.warning(f"Cannot parse date '{text}'.", warn)
    return None


def date_to_str(date):
    if date:
        return date.strftime("%Y-%m-%d")
    return "<UNDEFINED>"


def _reset_options_time():
    global _time_formats
    _time_formats = ["%H:%M", "%H.%M", "%H"]


def _set_options_time(config):
    formats = utils.expect_list_of_str(config.get("format"))
    for f in formats:
        _time_formats.append(f)


def parse_time(text, warn=True):
    clean_text = text.strip().casefold()
    for f in _time_formats:
        with suppress(ValueError):
            return datetime.datetime.strptime(clean_text, f).time()
    log.warning(f"Cannot parse time '{text}'.", warn)
    return None


def time_to_str(time):
    if time:
        return time.strftime("%H:%M")
    return "<UNDEFINED>"


def _reset_options_weekday():
    WeekdayTranslator.reset_options()


def _set_options_weekday(config, name):
    WeekdayTranslator.set_options(config.get("names") or {}, name)


def parse_weekday(text, warn=True):
    clean_text = str(text).strip().casefold()
    try:
        return WeekdayTranslator.get_number_by_name(clean_text)
    except KeyError:
        log.warning(f"Cannot parse weekday '{text}'.", warn)
        return None


def weekday_to_str(weekday):
    if weekday:
        name = WeekdayTranslator.get_name_by_number(weekday)
        return name or str(weekday)
    return "<UNDEFINED>"


reset_options()



from PyQt5.QtWidgets import (
    QDialog,
    QTabWidget,
    QVBoxLayout,
)

from meety.gui.info_dialog.about import TabAbout
from meety.gui.info_dialog.files import TabFiles
from meety.gui.info_dialog.log import TabLog
from meety.gui.info_dialog.meetings import TabMeetings
from meety.gui.info_dialog.system import TabSystem


class InfoDialog(QDialog):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Info")
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = QVBoxLayout()
        tabs = QTabWidget()
        tabs.setContentsMargins(0, 0, 0, 0)
        tabs.addTab(TabAbout(), "&About")
        tabs.addTab(TabSystem(), "&System")
        tabs.addTab(TabFiles(), "&Files")
        tabs.addTab(TabMeetings(), "&Meetings")
        tabs.addTab(TabLog(), "&Log")
        layout.addWidget(tabs)
        self.setLayout(layout)

    def _connect_widget_signals(self):
        pass



"""Windows implementation for system-dependent functionality."""

import os
import subprocess

icon_resolution = 256
icon_extension = "ico"
icon_path = None
desktop_extension = None
desktop_path = None
executable = "meety.exe"


def frozen_script_and_executable(path):
    path_to_exe = os.path.join(path, executable)
    return {
        # This is a hack to allow to create a shortcut
        # to an executable via `pyshortcuts`
        "executable": f"%%comspec /K start '' '{path_to_exe}' & exit & rem",
        "script": "src/gui.py",
    }


def is_installed(app):
    """Determine whether program `app` can be found."""
    result = subprocess.run(
        ['where', app],
        capture_output=True,
        text=True,
    )
    return result.returncode == 0


def update_shortcut_database():
    pass



import os
import shutil

from PyQt5.QtWidgets import (
    QFileDialog,
    QLabel,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from meety import (
    io,
    resources,
)
from meety.gui import app


def TEMPLATE_CREATED_MSG(target):
    return f'''<b>Template file created!</b> Next, please
    <ol>
        <li>edit the file <a href="{target}">{target}</a>
        according to your needs and</li>

        <li>reload the changed specification
        <ul>
            <li>via the context menu or</li>
            <li>by restarting the application.</li>
        </ul>
        </li>
    </ol>
    '''


class TabFiles(QWidget):
    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = QVBoxLayout()
        layout.setSpacing(3)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.addWidget(self._create_status())
        layout.addStretch(1)
        layout.addWidget(self._create_template_label())
        layout.addWidget(self._create_template_button())
        self.setLayout(layout)

    def _create_template_label(self):
        self._template_label = QLabel("<hr>")
        self._template_label.setWordWrap(True)
        return self._template_label

    def _create_template_button(self):
        self._template_button = QPushButton(
            "&Create specification from template"
        )
        self._template_button.setToolTip(
            "You can create a new meeting specification "
            "by editing a template. Click this button to "
            "save the template in a file."
        )
        self._template_button.setProperty("class", "create_template")
        return self._template_button

    def _create_status(self):
        self._status = QLabel(self.status)
        self._status.setOpenExternalLinks(True)
        return self._status

    def _connect_widget_signals(self):
        self._template_button.clicked.connect(self._create_template)

    def _create_template(self):
        source = resources.get_spec_path("example.yaml")
        target = self._ask_for_filename()
        if not target:
            return

        msg = QMessageBox()
        try:
            shutil.copy(source, target)
        except Exception as e:
            msg.setIcon(QMessageBox.Warning)
            msg.setText("Error on template creation.")
            msg.setDetailedText(str(e))
        else:
            app.add_meeting_file(target)
            msg.setIcon(QMessageBox.Information)
            msg.setText(TEMPLATE_CREATED_MSG(target))
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()

    def _ask_for_filename(self):
        filename = QFileDialog.getSaveFileName(
            parent=QFileDialog(),
            caption="New meeting specification",
            directory=os.path.expanduser("~"),
            filter="yaml (*.yaml)"
        )[0]
        return filename

    @property
    def status(self):
        return f"{self.status_directories}<br/>{self.status_files}"

    @property
    def status_directories(self):
        status = "<b>Considered directories:</b> "
        paths = app.get_active_directories()
        if paths:
            stats = [self.status_directory(p) for p in paths]
            status += io.html.unordered_list(stats)
        else:
            status += "no directories"
        return status

    def status_directory(self, path):
        return f'<a href="{path}">{path}</a>'

    @property
    def status_files(self):
        status = "<b>Considered files:</b> "
        loaded_paths = app.get_loaded_paths()
        paths = loaded_paths.all_paths
        if paths:
            stats = [
                self.status_file(p, loaded_paths.status(p))
                for p in paths
            ]
            status += io.html.unordered_list(stats)
        else:
            status += "no files"
        return status

    def status_file(self, path, info):
        link = f'<a href="{path}">{path}</a>'
        info = f'<i>({info})</i>'
        return f"{link} {info}"



import os
import shutil

import appdirs
import yaml

from meety import (
    connect,
    io,
    meetings,
    resources,
)
from meety.cli import styles as cli_styles
from meety.gui import styles as gui_styles
from meety.logging import log


class Config:
    """Configuration management for the application.

    The configuration may depend on multiple YAML files, loaded in
    a specific order. First, the internal "default" configuration,
    provided by the installed app, is loaded. Then, the user
    configuration is loaded, if found in the directory.

    Both these configuration files can activate and deactivate
    additional (internal) configurations: language-specific aspects
    and colour schemes for example.

    Later settings generally overwrite previous settings (some
    exceptions, e.g. for handlers, are described in the
    documentation).
    """

    PROGRAM = ""

    @classmethod
    def _ensure_user_config(cls, source, target):
        """Copy template to user config directory if it does not
        contain a configuration file.
        """
        if not os.path.isfile(target):
            log.info(
                f"Copying user configuration template to {target}."
            )
            shutil.copy(source, target)

    def __init__(self):
        """Prepare loading of settings. They will become effective
        only after `reload()` and `apply()`.
        """
        self._init_default_config()
        self._init_user_config()
        self._additional_configs = []

    def _init_default_config(self):
        self._defaults = PartialConfig(
            description="default",
            path=resources.get_config_path("defaults.yaml"),
            required=True
        )

    def _init_user_config(self):
        alternative_user_config = os.environ.get("MEETY_USER_CONFIG")
        if alternative_user_config is None:
            self._init_user_config_standard()
        else:
            self._init_user_config_alternative(alternative_user_config)

    def _init_user_config_standard(self):
        Config._ensure_user_config(
            resources.get_config_path("user.yaml"),
            user_config_path("config.yaml")
        )
        self._user = PartialConfig(
            description="user",
            path=user_config_path("config.yaml")
        )

    def _init_user_config_alternative(self, path):
        abspath = os.path.abspath(path)
        log.info(f"Using alternative user configuration file '{abspath}'.")
        self._user = PartialConfig(
            description="user",
            path=user_config_path(abspath)
        )

    def reload(self):
        """(Re-)Load the configuration from the predefined files."""
        self.defaults.load()
        self.user.load()
        self._load_additional_configs(
            self._determine_additional_configs()
        )

    def _determine_additional_configs(self):
        """Determine additional partial configurations based on
        the settings in the default and user configurations.
        """
        additionals = self.defaults.additional_configs.copy()
        additionals.update(self.user.additional_configs)
        return additionals

    def apply(self):
        """Apply all registered partial configurations."""
        self.defaults.apply()
        self.user.apply()
        for additional_config in self.additional_configs:
            additional_config.apply()

    def _load_additional_configs(self, additionals):
        for name, status in additionals.items():
            if io.str_is_true(status):
                self._load_additional_config(name)

    def _load_additional_config(self, name):
        filename = f"{name}.yaml"
        config = PartialConfig(
            description=name,
            path=resources.get_config_path(filename)
        )
        config.load()
        self._additional_configs.append(config)

    @property
    def defaults(self):
        """Return partial configuration for application defaults."""
        return self._defaults

    @property
    def user(self):
        """Return partial configuration by user."""
        return self._user

    @property
    def additional_configs(self):
        """Return additional partial configurations."""
        return self._additional_configs

    @property
    def all_partial_configs(self):
        """Return all partial configurations (defaults, user,
        additional).
        """
        return [self.defaults, self.user] + self.additional_configs

    def __str__(self):
        return "\n".join([str(c) for c in self.all_partial_configs])


class PartialConfig:
    """Manages the configuration provided by a single YAML file."""

    def __init__(self, description, path, required=False):
        """Prepare configuration for loading and application."""
        self._description = description
        self._path = path
        self._data = {}
        self._required = required

    @property
    def description(self):
        return self._description

    @property
    def data(self):
        return self._data

    @property
    def additional_configs(self):
        return io.expect_dict(self.data.get("additional_config"))

    @property
    def cli(self):
        return io.expect_dict(self.data.get("cli"))

    @property
    def cli_styles(self):
        return io.expect_dict(self.cli.get("styles"))

    @property
    def gui(self):
        return io.expect_dict(self.data.get("gui"))

    @property
    def gui_styles(self):
        return io.expect_dict(self.gui.get("styles"))

    @property
    def meetings(self):
        return io.expect_dict(self.data.get("meetings"))

    @property
    def connect(self):
        return io.expect_dict(self.data.get("connect"))

    def load(self):
        """Store loaded dictionary or empty dictionary, on failure."""
        self._data = self._load()

    def _load(self):
        status = "FAILED"
        config = {}
        try:
            with open(self._path) as cfile:
                config = yaml.load(cfile, Loader=yaml.BaseLoader)
                assert isinstance(config, dict)
        except FileNotFoundError:
            if self._required:
                log.warning(f"Cannot find configuration '{self.description}'.")
            else:
                log.warning(f"No configuration '{self.description}'.")
        except Exception as e:
            log.warning(f"Cannot read {self.description} configuration.")
            log.exception(e)
        else:
            status = "successful"
        finally:
            log.info(f"Load configuration '{self.description}': {status}")
            return config

    def app(self, app_name):
        return self.data.get(app_name, {})

    def args(self, app_name):
        """Return predefined list of command-line arguments for argparser."""
        arguments = self.app(app_name).get("args", [])
        split_args = []
        for argument in arguments:
            split_args.extend(argument.split(" "))
        return split_args

    def apply(self):
        """Apply configuration, pass information to the submodules."""
        connect.set_options(self.connect)
        meetings.set_options(self.meetings, self.description)
        cli_styles.set_options(self.cli_styles)
        gui_styles.set_options(self.gui_styles)

    def __str__(self):
        """Return description and textual representation of the dictionary."""
        return f"{self.description}:\n{self.data}"


def user_config_path(filename):
    """Return the path of the user file with the given name."""
    dir_path = appdirs.user_config_dir(Config.PROGRAM)
    os.makedirs(dir_path, exist_ok=True)
    return os.path.join(dir_path, filename)


def open_user_config(filename):
    """Return stream to the user file with the given name."""
    path = user_config_path(filename)
    return open(path)



"""Linux implementation for system-dependent functionality."""

import os
import subprocess

icon_resolution = 256
icon_extension = "png"
icon_path = os.path.expanduser("~/.icons")
desktop_extension = "desktop"
desktop_path = os.path.expanduser("~/.local/share/applications")
executable = "meety"
frozen_script_and_executable = None


def is_installed(app):
    """Determine whether program `app` can be found."""
    result = subprocess.run(
        ['which', app],
        capture_output=True,
        text=True,
    )
    return result.returncode == 0


def update_shortcut_database():
    subprocess.run(["update-desktop-database"])



from PyQt5 import QtCore
from PyQt5.QtWidgets import (
    QCheckBox,
    QLabel,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from meety.system import create_shortcut

EXPLANATION = """<b>Create shortcuts to <i>Meety</i> (GUI application).</b>
<br/><br/>
Whether shortcuts are added to your desktop and/or your
startmenu depends on your choice below <i>and on your system</i>.
"""


class TabSystem(QWidget):
    def __init__(self):
        super().__init__()
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = QVBoxLayout()
        layout.setSpacing(3)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.addWidget(self._create_explanation())
        layout.addWidget(self._create_desktop_option())
        layout.addWidget(self._create_startmenu_option())
        layout.addWidget(self._create_shortcut_button())
        layout.addStretch(1)
        self.setLayout(layout)

    def _create_explanation(self):
        self._explanation = QLabel(EXPLANATION)
        self._explanation.setWordWrap(True)
        self._explanation.setAlignment(QtCore.Qt.AlignTop)
        return self._explanation

    def _create_desktop_option(self):
        self._desktop_option = QCheckBox("on &desktop")
        self._desktop_option.setChecked(True)
        return self._desktop_option

    def _create_startmenu_option(self):
        self._startmenu_option = QCheckBox("on &startmenu")
        self._startmenu_option.setChecked(True)
        return self._startmenu_option

    def _create_shortcut_button(self):
        self._shortcut_button = QPushButton("&create")
        return self._shortcut_button

    def _connect_widget_signals(self):
        self._shortcut_button.clicked.connect(self._create_shortcut)

    def _create_shortcut(self):
        desktop = self._desktop_option.isChecked()
        startmenu = self._startmenu_option.isChecked()
        msg = QMessageBox()
        try:
            result = create_shortcut(desktop, startmenu)
        except Exception as e:
            msg.setIcon(QMessageBox.Warning)
            msg.setText("Error on shortcut creation.")
            msg.setDetailedText(str(e))
        else:
            if result is True:
                msg.setIcon(QMessageBox.Information)
                msg.setText("Shortcut(s) created.")
            else:
                msg.setIcon(QMessageBox.Warning)
                msg.setText(f"Shortcut(s) <i>not</i> created: {result}")
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()



from urllib.parse import urlparse


def expect_dict(datum, otherwise={}):
    if isinstance(datum, dict):
        return datum
    return otherwise


def expect_list(datum, otherwise=[]):
    if isinstance(datum, list):
        return datum
    return otherwise


def expect_list_of_str(datum, otherwise=[]):
    if isinstance(datum, str):
        return [datum]
    elif isinstance(datum, list):
        return [str(d) for d in datum]
    return otherwise


def str_is_true(text):
    clean_text = text.strip().casefold()
    return clean_text in ["yes", "true", "on"]


def str_is_false(text):
    return not str_is_true(text)


def ensure_between(value, at_least, at_most, on_wrong_type):
    """Ensure that numerical value is in given bounds."""
    try:
        if value < at_least:
            return at_least
        if value > at_most:
            return at_most
        return value
    except TypeError:
        return on_wrong_type


def is_url(text):
    try:
        url = urlparse(text, allow_fragments=False)
    except ValueError:
        return False
    else:
        return bool(url.netloc)


def replace_words_by(text, words, replacement):
    """
    Replace all given words by the replacement text.

    >>> replace_words_by('Monday and Friday', [' and '], ', ')
    'Monday, Friday'
    >>> replace_words_by('Monday und Friday', [' and '], ', ')
    'Monday und Friday'
    """
    for w in words:
        text = text.replace(w, replacement)
    return text



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QListWidgetItem


class MeetingItem(QListWidgetItem):
    def __init__(self, rmeeting):
        self._rmeeting = rmeeting
        super().__init__(self.name)
        self._init_style()
        self._init_meta_data()

    def _init_style(self):
        pass

    def _init_meta_data(self):
        for role, value in [
            (Qt.DisplayRole, self.name),
            (Qt.UserRole, self.rated_meeting),
            (Qt.ToolTipRole, self.info),
        ]:
            self.setData(role, value)

    @property
    def rating(self):
        return self._rmeeting.rating

    @property
    def meeting(self):
        return self._rmeeting.meeting

    @property
    def rated_meeting(self):
        return self._rmeeting

    @property
    def name(self):
        return self.meeting.name

    @property
    def info(self):
        return self.meeting.data_to_str()



"""Configuration for the graphical user interface."""

from meety.io import utils
from meety.logging import log

STYLES = {}
ATTRIBUTES = [
    "show",
    "size",
    "color",
    "background",
    "selected",
    "attrs",
]
DEFAULTS = {
    "show": "yes",
    "size": "18",
    "color": "black",
    "background": "white",
    "selected": "white",
    "attrs": [],
}


def set_options(config):
    if not isinstance(config, dict):
        log.expected("dict", "cli", config)
    else:
        for k, v in config.items():
            _add_style(k, v)


def _add_style(style_name, style_def):
    if not isinstance(style_def, dict):
        log.expected("dict", "styles", style_def)
        return
    STYLES[style_name] = {key: style_def.get(key) for key in ATTRIBUTES}


def get_style_value(style_name, value):
    return STYLES.get(style_name).get(value) or DEFAULTS.get(value)


def get_show(style_name="others"):
    return utils.str_is_true(get_style_value(style_name, "show"))


def get_color(style_name="others"):
    return get_style_value(style_name, "color")


def get_background(style_name="others"):
    return get_style_value(style_name, "background")


def get_selected(style_name="others"):
    return get_style_value(style_name, "selected")


def get_attrs(style_name="others"):
    return get_style_value(style_name, "attrs")


def get_font_size(style_name="others"):
    return int(get_style_value(style_name, "size"))


def is_bold(style_name="others"):
    return "bold" in get_attrs(style_name)


def get_border_color():
    return "#aaa"



from PyQt5.QtCore import (
    QPoint,
    Qt,
    pyqtSignal,
)
from PyQt5.QtWidgets import (
    QApplication,
    QVBoxLayout,
    QWidget,
)

from meety import resources
from meety.gui import static as gui_static
from meety.gui.main_window.meeting_list import Meetings
from meety.gui.main_window.search import SearchWidget
from meety.gui.main_window.status import StatusWidget
from meety.io.utils import ensure_between


class MainWindow(QWidget):
    query_changed = pyqtSignal(str, name="query_changed")
    rating_mode_toggled = pyqtSignal(bool, name="rating_mode_toggled")
    meeting_chosen = pyqtSignal(object, name="meeting_chosen")
    handler_chosen = pyqtSignal(object, name="handler_chosen")
    reload_requested = pyqtSignal(name="reload_requested")

    def __init__(self, width, height, xpos=0.5, ypos=0.5):
        super().__init__()
        self.resize(width, height)
        self.move(relative_screen_position(xpos, ypos, width, height))
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()

    def _init_style(self):
        with resources.open_text(gui_static, "stylesheet.css") as css:
            self.setStyleSheet(css.read())

    def _add_widgets(self):
        layout = self._create_layout()
        layout.addWidget(self._create_search())
        layout.addWidget(self._create_meetings())
        layout.addWidget(self._create_status())
        self.setLayout(layout)

    def _create_layout(self):
        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        return layout

    def _create_search(self):
        self._search = SearchWidget()
        return self._search

    def _create_meetings(self):
        self._meetings = Meetings()
        return self._meetings

    def _create_status(self):
        self._status = StatusWidget()
        return self._status

    def _connect_widget_signals(self):
        self._search.textChanged.connect(self.query_changed.emit)
        self._meetings.rating_mode_toggled.connect(
            self.rating_mode_toggled.emit
        )
        self._meetings.handler_chosen.connect(self.handler_chosen.emit)
        self._meetings.meeting_chosen.connect(self.meeting_chosen.emit)
        self._meetings.reload_requested.connect(self.reload_requested.emit)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        self.on_key(event)

    def on_key(self, event):
        key = event.key()
        if key == Qt.Key_Down:
            self._meetings.move_selection_down()
        elif key == Qt.Key_Up:
            self._meetings.move_selection_up()
        elif key == Qt.Key_Return:
            self.meeting_chosen.emit(self._meetings.current_meeting)

    def update_rated_meetings(self, rated_meetings):
        self._meetings.update_rated_meetings(rated_meetings)

    def set_meetings_name(self, name):
        self._meetings.set_name(name)

    def notify_quickly(self, text):
        self._status.notify_quickly(text)

    def notify(self, text, tooltip=""):
        self._status.notify(text, tooltip)

    def warn(self, text, tooltip=""):
        self._status.warn(text, tooltip)

    def clear_notifications(self):
        self._status.clear_notifications()


def relative_screen_position(rel_xpos, rel_ypos, width, height):
    screen = QApplication.desktop().screenGeometry()
    screen.adjust(0, 0, -width, -height)
    rel_xpos = ensure_between(
        value=rel_xpos,
        at_least=0.0,
        at_most=1.0,
        on_wrong_type=0.5
    )
    rel_ypos = ensure_between(
        value=rel_ypos,
        at_least=0.0,
        at_most=1.0,
        on_wrong_type=0.5
    )
    return QPoint(
        rel_xpos * screen.width(),
        rel_ypos * screen.height()
    )



from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon

from meety import resources


def get_icon(name, sizes=[256]):
    icon = QIcon()
    for size in sizes:
        filename = f"{name}-{size}.png"
        _add_file_to_icon(icon, filename, size)
    return icon


def _add_file_to_icon(icon, filename, size):
    path = resources.get_icon_path(filename)
    icon.addFile(str(path), QSize(size, size))



"""Configuration for the command-line interface."""

import colorama
from termcolor import colored

from meety.logging import log

colorama.init()
STYLES = {}


def set_options(config):
    if not isinstance(config, dict):
        log.expected("dict", "cli", config)
    else:
        for k, v in config.items():
            _add_style(k, v)


def _add_style(style_name, style_def):
    if not isinstance(style_def, dict):
        log.expected("dict", "styles", style_def)
        return

    STYLES[style_name] = {
        key: style_def.get(key)
        for key in ["color", "background", "attrs"]
    }


def style(text, style_name="others"):
    style = STYLES.get(style_name) or {}
    return colored(
        text,
        style.get("color"),
        style.get("background"),
        attrs=style.get("attrs")
    )


def bf(text):
    return colored(text, attrs=["bold"])


def bf_str(datum):
    return bf(str(datum))



from PyQt5.QtCore import (
    QEvent,
    QMargins,
    QRect,
    Qt,
    pyqtSignal,
)
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QFont,
    QPen,
)
from PyQt5.QtWidgets import (
    QApplication,
    QItemDelegate,
    QStyle,
    QStyleOptionButton,
)

from meety.gui import styles


class ItemDelegate(QItemDelegate):
    """Draw meeting items according to their match status and the GUI
    configuration.
    """

    context_menu_requested = pyqtSignal(object, name="context_menu_requested")

    def __init__(self, parent=None, *args):
        QItemDelegate.__init__(self, parent, *args)
        self._context_menu_button_rects = {}

    def paint(self, painter, option, index):
        painter.save()
        self._draw_background(painter, option, index)
        self._draw_border(painter, option)
        self._draw_context_menu_button(painter, option, index)
        self._draw_text(painter, option, index)
        painter.restore()

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseButtonRelease:
            rect = self._context_menu_button_rects[index]
            if rect.contains(event.pos()):
                self.context_menu_requested.emit(event)
                return True
        return super().editorEvent(event, model, option, index)

    def sizeHint(self, option, index):
        orig_size = super().sizeHint(option, index)
        return orig_size.grownBy(QMargins(5, 10, 5, 15))

    @classmethod
    def _draw_background(cls, painter, option, index):
        painter.setPen(QPen(Qt.NoPen))
        cls._set_bg_color(painter, option, index)
        painter.drawRect(option.rect)

    @classmethod
    def _set_bg_color(cls, painter, option, index):
        rmeeting = index.data(Qt.UserRole)
        if option.state & QStyle.State_Selected:
            color = styles.get_selected(rmeeting.match_style)
        else:
            color = styles.get_background(rmeeting.match_style)
        painter.setBrush(QBrush(QColor(color)))

    @classmethod
    def _draw_border(cls, painter, option):
        painter.setPen(QPen(QColor(styles.get_border_color())))
        painter.drawLine(
            option.rect.bottomLeft(),
            option.rect.bottomRight()
        )

    def _draw_context_menu_button(self, painter, option, index):
        style = QStyleOptionButton()
        style.features = QStyleOptionButton.Flat
        style.rect = self._context_menu_button_rect(option, index)
        style.text = "☰"
        style.state = QStyle.State_Enabled

        painter.setPen(QColor("blue"))
        QApplication.style().drawControl(
            QStyle.CE_PushButton,
            style,
            painter
        )

    def _context_menu_button_rect(self, option, index):
        anchor = option.rect.bottomRight()
        margins = QMargins(3, 3, 3, 3)
        rect = QRect(0, 0, 30, 30)
        rect.moveBottomRight(anchor)
        self._context_menu_button_rects[index] = rect
        return rect.marginsRemoved(margins)

    @classmethod
    def _draw_text(cls, painter, option, index):
        rmeeting = index.data(Qt.UserRole)
        rect = option.rect.marginsRemoved(QMargins(5, 5, 5, 0))
        cls._draw_text_name(painter, rect, rmeeting)
        cls._draw_text_info(painter, rect, rmeeting)

    @classmethod
    def _draw_text_name(cls, painter, rect, rmeeting):
        title = rmeeting.meeting.name
        cls._set_font_properties(painter, rmeeting.match_style)
        painter.drawText(rect, Qt.AlignLeft, title)

    @classmethod
    def _draw_text_info(cls, painter, rect, rmeeting):
        info = rmeeting.match_preference_details
        if info and styles.get_show("info"):
            cls._set_font_properties(painter, "info")
            painter.drawText(rect, Qt.AlignLeft | Qt.AlignBottom, info)

    @classmethod
    def _set_font_properties(cls, painter, style):
        cls._set_font_color(painter, style)
        cls._set_font(painter, style)

    @classmethod
    def _set_font_color(cls, painter, style):
        color = styles.get_color(style)
        painter.setPen(QPen(QColor(color)))

    @classmethod
    def _set_font(cls, painter, style):
        font = QFont()
        font.setPixelSize(styles.get_font_size(style))
        font.setBold(styles.is_bold(style))
        painter.setFont(font)



import pyperclip

from meety.io import (
    datetime,
    html,
)
from meety.io.actions import registry
from meety.io.utils import *


def reset_options():
    datetime.reset_options()
    registry.reset_options()


def set_options(config, name):
    datetime.set_options(config.get("preferences") or {}, name)
    registry.set_options(config.get("attributes") or [])


def split_on_semicolon_or_comma(text):
    if ";" in text:
        return text.split(";")
    return text.split(",")


def parse_first_possible(parser, tokens):
    for token in tokens:
        result = parser(token, False)
        if result and result.is_valid:
            return result
    return None


def process(data):
    data = process_recursively(data, _casefold_keys)
    for action in registry.get_actions():
        data = process_recursively(data, action)
    return data


def process_recursively(value, func, level=1):
    if isinstance(value, list):
        return func(
            [
                process_recursively(v, func, level+1)
                for v in value
            ],
            level
        )
    if isinstance(value, dict):
        return func(
            {
                k: process_recursively(v, func, level+1)
                for (k, v) in value.items()
            },
            level
        )
    else:
        return func(value, level)


def _casefold_keys(data, level):
    if isinstance(data, dict):
        return {
            k.casefold(): v
            for (k, v) in data.items()
        }
    return data


def copy_to_clipboard(text):
    pyperclip.copy(text)



import re

from meety.io.actions import registry
from meety.io.actions.action import Action


class ReplaceAction(Action):
    """Replace the value for the attribute with `key` if the value
    matches the expression and the substitute is defined by the
    dictionary attributes.
    """

    name = "replace"

    def __init__(self, key, expression, substitute):
        super().__init__()
        self._key = key
        self._pattern = re.compile(expression)
        self._substitute = substitute

    def is_applicable(self, data, current_level):
        return (
            super().is_applicable(data, current_level)
            and self._key in data.keys()
        )

    def _apply(self, data):
        value = data[self._key]
        template = self._pattern.sub(self._substitute, value)
        try:
            new_value = template.format_map(data)
        except KeyError:
            new_value = value
        finally:
            data[self._key] = new_value
        return data

    @classmethod
    def create(cls, data):
        input_ = data["input"]
        expression = data["expression"]
        substitute = data["substitute"]
        return ReplaceAction(input_, expression, substitute)

    @classmethod
    def register(cls):
        registry.register_action_class("replace", ReplaceAction)



"""System-dependent functionality."""

import os
import pathlib
import platform
import shutil
import sys

from meety import resources
from meety.logging import log

DESCRIPTION = """Meety - quickly start online meetings from YAML."""

SYSTEM = platform.system()
_running = None
log.debug(f"Your system is '{SYSTEM}'")
if SYSTEM == "Linux":
    from meety.system import linux
    _running = linux
elif SYSTEM == "Darwin":
    from meety.system import darwin
    _running = darwin
elif SYSTEM == "Windows":
    from meety.system import windows
    _running = windows
else:
    log.warning(f"System '{SYSTEM}' is unknown.")


def icon_resolution():
    return _running.icon_resolution


def icon_extension():
    return _running.icon_extension


def desktop_extension():
    return _running.desktop_extension


def desktop_path(filename):
    return os.path.join(_running.desktop_path, filename)


def icon_path(filename):
    return os.path.join(_running.icon_path, filename)


def executable():
    return _running.executable


def satisfies(system, installed):
    """Tests whether the app runs on the given system and whether the
    given program is installed."""
    return all([
        not system or SYSTEM.lower() == system.lower(),
        not installed or _running.is_installed(installed),
    ])


def create_shortcut(desktop=True, startmenu=True):
    filename = f"meety-{icon_resolution()}.{icon_extension()}"
    icon_path = resources.get_icon_path(filename)
    log.debug(f"Creating shortcut with icon '{icon_path}'.")

    try:
        msg = create_shortcut_via_pyshortcuts(icon_path, desktop, startmenu)
    except ImportError:
        log.warning("Module 'pyshortcuts' is not available!")
        msg = try_to_copy_desktop_file(filename)

    if msg is True:
        _running.update_shortcut_database()
    return msg


def create_shortcut_via_pyshortcuts(icon_path, desktop=True, startmenu=True):
    from pyshortcuts import make_shortcut

    target = script_and_executable()
    if target is None:
        log.error("Script/executable is not defined. Cannot create shortcut.")
        return False
    else:
        log.debug(f"Script/executable = '{target}'")
    make_shortcut(
        **target,
        name="Meety",
        description=DESCRIPTION,
        icon=icon_path,
        terminal=False,
        desktop=desktop,
        startmenu=startmenu,
    )
    return True


def try_to_copy_desktop_file(filename):
    log.info("Try to copy desktop/icon file.")
    try:
        copy_desktop_file()
        copy_icon(filename)
    except Exception as e:
        return str(e)
    else:
        return True


def copy_desktop_file():
    filename = f"meety.{desktop_extension()}"
    source = resources.get_desktop_path(filename)
    target = desktop_path(filename)
    shutil.copy(source, target)


def copy_icon(filename):
    source = resources.get_icon_path(filename)
    target = icon_path(filename)
    shutil.copy(source, target)


def script_and_executable():
    if getattr(sys, "frozen", False):
        log.debug("Application is 'frozen'.")
        raise Exception(
            "Shortcut creation for application bundles is not yet supported. "
            "Sorry!"
        )
        # TODO: Support application bundles
        # path = sys._MEIPASS
        # return _running.frozen_script_and_executable(path)
    else:
        log.debug("Application is _not_ 'frozen'.")
        system_path = os.path.dirname(os.path.abspath(__file__))
        path = pathlib.Path(system_path)
        base_path = path.parent.parent
        return {
            "script": os.path.join(base_path, "gui.py")
        }



"""Some utility functions for the command-line interface."""

from meety.cli import styles


def print_indexed_names(items, name_func=styles.bf_str):
    max_index = len(items)
    digits = len(str(max_index))
    for index, item in enumerate(items, 1):
        name = name_func(item)
        print(f"  [{index:{digits}}] {name}")


def input_choice(valid_answers, default_answer, tries=3):
    for count in range(0, tries):
        if count > 0:
            print("Try again: ", end="")
        answer = input_choice_once(valid_answers, default_answer)
        if answer:
            return answer
        print("Unrecognised answer.", end=" ")
    return default_answer


def input_choice_once(valid_answers, default_answer):
    answer = input()
    if answer in valid_answers:
        return answer
    elif answer.strip() == "":
        return default_answer
    else:
        return None


def index_choice(start_index, end_index, default_answer, tries=3):
    valid_indices = range(start_index, end_index + 1)
    valid_answers = [str(i) for i in valid_indices]
    answer = input_choice(valid_answers, default_answer, tries)
    return int(answer)


def yes_no_choice(default_answer='n', tries=3):
    return input_choice(
        ['y', 'Y', 'n', 'N'],
        default_answer,
        tries
    ).lower()



"""The command-line interface program."""

from datetime import datetime

from meety import connect
from meety.cli import styles as styles
from meety.cli.args import add_argparser_arguments  # USED
from meety.cli.utils import (
    index_choice,
    print_indexed_names,
    yes_no_choice,
)
from meety.io import copy_to_clipboard
from meety.logging import log
from meety.meetings.rating import Ratings

name = "cli"


def run(args, loader):
    meetings = loader.meetings
    if args.list_connection_handlers:
        print(connect.handlers)
    elif args.list_meetings:
        _list_meeting_details(meetings)
    else:
        matches = _rate_and_match(args, meetings)
        if args.list_ratings:
            _show_ratings(matches)
        else:
            _show_and_connect(args, matches)


def _show_ratings(matches):
    print("Ratings (Q|T|x = matches query|time|not):")
    print("\n".join([f"  - {rm.debug_info()}" for rm in matches]))


def _rate_and_match(args, meetings):
    rating = Ratings(meetings)
    when = _get_datetime(args.datetime)
    rating.rate(args.query, when)
    return _get_matches(args, rating)


def _show_and_connect(args, matches):
    _show_meetings(matches)
    if not args.show_only:
        _choose_and_connect(args, matches)


def _choose_and_connect(args, matches):
    meeting = _choose_meeting(matches)
    if meeting:
        _try_to_connect_to_meeting(args, meeting)
    else:
        print("No meeting found. Goodbye!")


def _list_meeting_details(meetings):
    print("Meeting details:")
    for m in meetings:
        print(m.debug_info())


def _get_datetime(text):
    if not text:
        return datetime.now()

    when = _parse_datetime_argument(text)
    if not when:
        log.expected("format yyyy-mm-ddTHH:MM", "datetime", text)
        return None


def _parse_datetime_argument(text):
    try:
        when = datetime.strptime(text, "%Y-%m-%dT%H:%M")
    except ValueError:
        return None
    else:
        return when


def _get_matches(args, rating):
    if args.all:
        return rating.get_all()
    elif args.best:
        return rating.get_only_best()
    elif args.first:
        return rating.get_only_first()
    else:
        return rating.get_all_matching()


def _show_meetings(matches):
    count = len(matches)
    if count == 1:
        _show_single_meeting(matches[0])
    elif count > 1:
        _show_multiple_meetings(matches)


def _show_single_meeting(match):
    text = _print_rated_meeting(match)
    print(f"There's one matching meeting: {text}")


def _show_multiple_meetings(matches):
    print("There are multiple matches:")
    print_indexed_names(matches, _print_rated_meeting)
    print("Which one do you want to choose?", end=" ")


def _print_rated_meeting(rmeeting):
    name = str(rmeeting)
    style_name = rmeeting.match_style
    return styles.style(name, style_name)


def _choose_meeting(matches):
    count = len(matches)
    if count == 1:
        return matches[0].meeting
    if count > 1:
        index = index_choice(1, len(matches), "1", 3) - 1
        return matches[index].meeting


def _try_to_connect_to_meeting(args, meeting):
    if args.yes or _ask_to_connect():
        _connect_to_meeting(args, meeting)


def _connect_to_meeting(args, meeting):
    if args.copy_password:
        _copy_password(meeting)
    handler = _get_handler(meeting, args.choose_connection)
    if handler:
        _connect_with_handler(handler, args.test_run)


def _get_handler(meeting, choose):
    handlers = connect.applicable_handlers(meeting)
    log.debug(f"Handlers found: {handlers}")
    handler = _choose_handler(handlers, choose)
    log.debug(f"Handler chosen: {handler}")
    return handler


def _connect_with_handler(handler, test_run):
    connect.start(handler, test_run)
    if test_run:
        print(
            styles.bf("This is a test run. ")
            + "The following command would be used in a normal run."
            + f"\n\n{handler.cmd}"
        )


def _ask_to_connect():
    print("Really connect (y/N)?", end=" ")
    okay = yes_no_choice()
    print()
    if okay == "y":
        print("Fine, connecting ...")
        return True
    else:
        print("Okay, aborting. Goodbye!")
        return False


def _copy_password(meeting):
    password = meeting.data.get("password")
    if password:
        print("...Copied password to clipboard.")
        copy_to_clipboard(password)


def _choose_handler(handlers, choose_connection):
    if not handlers:
        print("Cannot find a connection handler!")
        return None
    if len(handlers) == 1 or not choose_connection:
        return handlers[0]

    return _choose_from_multiple_handlers(handlers)


def _choose_from_multiple_handlers(handlers):
    _print_multiple_handlers(handlers)
    return _ask_for_handler(handlers)


def _print_multiple_handlers(handlers):
    print("There are multiple connection handlers:")
    print_indexed_names([h.name for h in handlers])
    print("Which one do you want to choose?", end=" ")


def _ask_for_handler(handlers):
    index = index_choice(1, len(handlers), "1", 3) - 1
    return handlers[index]



import yaml

from meety.io import is_url
from meety.logging import log
from meety.meetings.preferences import (
    TimeMatchDescription,
    TimePreference,
)


class Meeting:
    """Meetings require a name (for reference), can have a password,
    connection information and multiple time preferences.

    The `data` dictionary can comprise further keys that might be
    ignored by Meety.
    """

    @classmethod
    def is_meeting_entry(cls, data):
        """Test whether the given data is a dictionary that contains
        an entry 'name'.
        """
        return (
            isinstance(data, dict)
            and "name" in [k.casefold() for k in data.keys()]
        )

    def __init__(self, data, clean_data):
        """Initialize meeting by dictionary `data`."""
        self._input_data = data,
        self._clean_data = clean_data
        self._name = self._clean_data.get("name")
        self._password = self._clean_data.get("password")

        self._init_preferences()
        self._init_query_entries()
        self._init_url_entries()

    def _init_preferences(self):
        self._preferences = []
        values = self._clean_data.get("prefer") or []
        if not isinstance(values, list):
            # wrap a single entry in a list:
            values = [values]
        for value in values:
            if isinstance(value, dict):
                pref = TimePreference.from_dict(value)
            elif isinstance(value, str):
                pref = TimePreference.from_str(value)
            else:
                log.expected(
                    "string or dictionary",
                    "time preference specification",
                    value
                )
                pref = None
            if pref and pref.is_nontrivial:
                self._preferences.append(pref)

    def _init_query_entries(self):
        self._query_entries = []
        # keep all simple values as query entries:
        for (key, value) in self._clean_data.items():
            if not isinstance(value, (list, dict)):
                self._query_entries.append(str(value).casefold())

    def _init_url_entries(self):
        self._urls = []
        for (key, value) in self._clean_data.items():
            if isinstance(value, str) and is_url(value):
                self._urls.append(value)

    @property
    def name(self):
        return self._name or "<UNNAMED>"

    @property
    def password(self):
        return self._password

    @property
    def preferences(self):
        return self._preferences

    @property
    def input_data(self):
        return self._input_data

    def input_data_to_str(self):
        return yaml.dump(
            self.input_data[0],
            default_flow_style=False
        )

    @property
    def data(self):
        return self._clean_data

    def data_to_str(self):
        return yaml.dump(
            self.data,
            default_flow_style=False
        )

    @property
    def query_entries(self):
        return self._query_entries

    @property
    def urls(self):
        return self._urls

    def match_time(self, when):
        """Tests whether any of the time preferences is matched for the
        given datetime `when`.
        """
        otherwise = [TimeMatchDescription.NOT_MATCHING]
        return max([p.match(when) for p in self.preferences] + otherwise)

    def match_query(self, query):
        """Computes the sum of matching query items."""
        clean_query = self._clean_query(query)
        return sum([
            any([(q in e) for e in self.query_entries])
            for q in clean_query
        ])

    @classmethod
    def _clean_query(cls, query):
        """Strip and casefold all query items."""
        return [
            non_empty.casefold() for non_empty in [
                q.strip() for q in query
            ]
            if len(non_empty) > 0
        ]

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.data == other.data
        return False

    def __hash__(self):
        kvs = sorted([
            str(k) + ":" + str(v)
            for k, v in self._clean_data.items()
        ], key=lambda kv: kv[0])
        return "\n".join(kvs).__hash__()

    def debug_info(self):
        return "\n".join([
            self._debug_info_base(),
            self._debug_info_prefs(),
            self._debug_info_others(),
            ""
        ])

    def _debug_info_base(self):
        p = len(self._preferences)
        return "\n".join([
            "Meeting:",
            f"  - name: {self.name}",
            f"  - password: {'yes <HIDDEN>' if self.password else 'no'}",
            f"  - preferences ({p})",
        ])

    def _debug_info_prefs(self):
        return "\n\n".join([
            f"{pref}" for pref in self._preferences
        ])

    def _debug_info_others(self):
        others = []
        for (k, v) in self.data.items():
            if k not in ["name", "password", "prefer"]:
                others.append(f"  - {k}: {v}")
        return "\n".join(others)



""" Time preferences that can comprise weekday, date and time specifications.
Each specification can be a scalar, a range or a list of ranges.
"""
import enum
from abc import (
    ABC,
    abstractmethod,
)
from datetime import (
    datetime,
    time,
    timedelta,
)
from functools import total_ordering

from meety import io
from meety.logging import log
from meety.meetings import utils


def reset_options():
    WeekdayInterval.reset_options()
    DateInterval.reset_options()
    TimeInterval.reset_options()
    io.reset_options()


def set_options(config, name):
    WeekdayInterval.set_options(config.get("weekday") or {})
    DateInterval.set_options(config.get("date") or {})
    TimeInterval.set_options(config.get("time") or {})
    io.datetime.set_options(config, name)


def to_datetime(d, t):
    try:
        return datetime.combine(d, t)
    except TypeError:
        return None


@total_ordering
class TimeMatchDescription(enum.Enum):
    NOT_MATCHING = 0
    NOT_DEFINED = 1
    MATCHING = 2
    WEEKDAY_MATCHING = 3
    DATE_MATCHING = 4
    TIME_MATCHING = 5

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def matches(description):
        return description not in [
            TimeMatchDescription.NOT_MATCHING,
            TimeMatchDescription.NOT_DEFINED,
        ]


class TimePreference:
    """Each time preference can refer to weekday, date and time conditions.
    The conditions can be lists of ranges.
    """

    @classmethod
    def from_dict(cls, data):

        def _parse_or_true(name, parser):
            att = data.get(name)
            if att:
                return parser(att)
            else:
                return TRUE_DISJUNCTION

        weekday = _parse_or_true("weekday", DisjunctiveIntervals.parse_weekday)
        date = _parse_or_true("date", DisjunctiveIntervals.parse_date)
        time = _parse_or_true("time", DisjunctiveIntervals.parse_time)
        return TimePreference(weekday=weekday, date=date, time=time)

    @classmethod
    def from_str(cls, text):

        def _parse_tokens(parser):
            return io.parse_first_possible(parser, tokens)

        tokens = io.split_on_semicolon_or_comma(text)
        weekday = _parse_tokens(DisjunctiveIntervals.parse_weekday)
        date = _parse_tokens(DisjunctiveIntervals.parse_date)
        time = _parse_tokens(DisjunctiveIntervals.parse_time)
        return TimePreference(weekday=weekday, date=date, time=time)

    def __init__(self, weekday, date, time):
        self._weekday = weekday
        self._date = date
        self._time = time

    def match(self, when):
        """Tests whether all conditions (weekday, time, date) are met."""
        subconditions = [
            p.is_satisfied(when, description)
            for name, p, description in self._named_and_weighted_preferences
        ]
        if TimeMatchDescription.NOT_MATCHING in subconditions:
            # not all conditions are matched
            return TimeMatchDescription.NOT_MATCHING
        else:
            # consider the strongest condition
            return max(subconditions)

    def __str__(self):
        prefs = [p for p in self._named_and_weighted_preferences if str(p[1])]
        if len(prefs) == 0:
            return "no preference"

        def describe(name, pref):
            return f"      {name}: {pref}"

        if len(prefs) == 1:
            return describe(prefs[0])
        else:
            return "\n".join([describe(n, p) for (n, p, w) in prefs])

    @property
    def is_nontrivial(self):
        p = list(self._named_and_weighted_preferences)
        return len(p) > 0

    @property
    def _named_and_weighted_preferences(self):
        if self._weekday:
            yield (
                "weekday",
                self._weekday,
                TimeMatchDescription.WEEKDAY_MATCHING
            )
        if self._date:
            yield (
                "date",
                self._date,
                TimeMatchDescription.DATE_MATCHING
            )
        if self._time:
            yield (
                "time",
                self._time,
                TimeMatchDescription.TIME_MATCHING
            )


class DisjunctiveIntervals:
    """A list of intervals that are considered as alternatives."""

    @classmethod
    def parse_weekday(cls, text, warn=True):
        return DisjunctiveIntervals(cls.parse(WeekdayInterval, text, warn))

    @classmethod
    def parse_date(cls, text, warn=True):
        return DisjunctiveIntervals(cls.parse(DateInterval, text, warn))

    @classmethod
    def parse_time(cls, text, warn=True):
        return DisjunctiveIntervals(cls.parse(TimeInterval, text, warn))

    @classmethod
    def parse(cls, interval_cls, text, warn=True):
        if text is None:
            return None
        clean_text = io.datetime.prepare_list_of_ranges(text)
        ranges = utils.list_of_ranges(clean_text)
        intervals = []
        for r in ranges:
            try:
                interval = interval_cls.parse(r, warn)
            except ValueError:
                log.warning(f"Cannot parse '{r}' in '{text}'.", warn)
            else:
                if interval and interval.is_valid:
                    intervals.append(interval)
        return intervals or None

    def __init__(self, intervals):
        self._intervals = intervals

    @property
    def is_valid(self):
        return self._intervals

    def is_satisfied(self, when, desc=TimeMatchDescription.NOT_MATCHING):
        """Tests whether the given `when` satisfies at least one
        interval or whether there is none.
        """
        if self._intervals is None:
            return TimeMatchDescription.NOT_MATCHING
        if not self._intervals:
            return TimeMatchDescription.NOT_DEFINED
        return desc if any(self.match_all(when)) \
            else TimeMatchDescription.NOT_MATCHING

    def match_all(self, when):
        """Apply matching test for `when` to each interval."""
        return map(lambda i: i.match(when), self._intervals)

    def __str__(self):
        """Return a comma-separated list of the intervals."""
        if not self._intervals:
            return "<UNDEFINED>"
        return ", ".join([str(i) for i in self._intervals])


FALSE_DISJUNCTION = DisjunctiveIntervals(None)
TRUE_DISJUNCTION = DisjunctiveIntervals([])


class AbstractInterval(ABC):
    """Common functionality for weekday, date and time intervals."""

    @classmethod
    def reset_options(cls):
        cls._start_delta = timedelta()
        cls._end_delta = timedelta()

    @classmethod
    def set_options(cls, config):
        if config:
            sdelta = config.get("start_delta")
            if sdelta:
                cls._start_delta = io.datetime.parse_timedelta(sdelta)
            edelta = config.get("end_delta")
            if edelta:
                cls._end_delta = io.datetime.parse_timedelta(edelta)

    @classmethod
    def parse(cls, entries, warn=True):
        clean_entries = [e.strip().casefold() for e in entries]
        stext, etext = utils.to_pair(clean_entries)
        start = cls._parse_entry(stext, warn)
        end = cls._parse_entry(etext, warn)
        return cls(start or end, end or start)

    @classmethod
    @abstractmethod
    def _parse_entry(cls, entry, warn=True):
        pass

    @classmethod
    @abstractmethod
    def _entry_to_string(cls, entry):
        pass

    def __init__(self, start, end):
        self._start = start
        self._end = end

    @property
    def is_valid(self):
        return None not in [self._start, self._end]

    @abstractmethod
    def match(self, when):
        pass

    def _is_between_datetime(self, start, when, end):
        try:
            lower = start + self._start_delta
            upper = end + self._end_delta
            return lower <= when <= upper
        except TypeError:
            return False

    def __str__(self):
        stext = self._entry_to_string(self._start)
        etext = self._entry_to_string(self._end)
        if stext == etext:
            return stext
        else:
            return "%s - %s" % (stext, etext)


class ConstantInterval(AbstractInterval):
    def __init__(self, value):
        super().__init__(None, None)
        self._value = value

    def match(self, when):
        return self._value

    @classmethod
    def _parse_entry(cls, entry, warn=True): pass

    @classmethod
    def _entry_to_string(cls, entry):
        return "<CONSTANT>"


FALSE_INTERVAL = ConstantInterval(False)
TRUE_INTERVAL = ConstantInterval(True)


class DateInterval(AbstractInterval):
    """An interval of dates.
    For instance, '2020-05-27' or '2020-05-27 - 2020-05-30'.
    """

    _start_delta = timedelta()
    _end_delta = timedelta()

    def __init__(self, start, end):
        super().__init__(start, end)

    def match(self, when):
        lower = to_datetime(self._start, time(0, 0))
        upper = to_datetime(self._end, time(23, 59))
        return self._is_between_datetime(lower, when, upper)

    @classmethod
    def _parse_entry(cls, text, warn=True):
        return io.datetime.parse_date(text, warn)

    @classmethod
    def _entry_to_string(cls, entry):
        return io.datetime.date_to_str(entry)


class TimeInterval(AbstractInterval):
    """An interval of times.
    For instance, '20:15' or '20:15 - 21:45'.
    """

    _start_delta = timedelta()
    _end_delta = timedelta()

    def __init__(self, start, end):
        super().__init__(start, end)

    def match(self, when):
        start = to_datetime(when.date(), self._start)
        end = to_datetime(when.date(), self._end)
        return self._is_between_datetime(start, when, end)

    @classmethod
    def _parse_entry(cls, text, warn):
        return io.datetime.parse_time(text, warn)

    @classmethod
    def _entry_to_string(cls, entry):
        return io.datetime.time_to_str(entry)


class WeekdayInterval(AbstractInterval):
    """An interval of weekdays.
    For instance,  'Sunday' or 'Monday - Friday'.
    """

    def __init__(self, start, end):
        super().__init__(start, end)

    def match(self, when):
        start = self._start
        day = when.isoweekday()
        end = self._end
        return self._is_between_days(start, day, end)

    def _is_between_days(self, start, when, end):
        try:
            return start <= when <= end
        except TypeError:
            return False

    @classmethod
    def _parse_entry(cls, text, warn=True):
        return io.datetime.parse_weekday(text, warn)

    @classmethod
    def _entry_to_string(cls, entry):
        return io.datetime.weekday_to_str(entry)



from PyQt5.QtWidgets import QLineEdit


class SearchWidget(QLineEdit):
    def __init__(self):
        super().__init__()
        super().setPlaceholderText("search")
        super().setProperty("class", "search")



import time

from PyQt5.QtCore import (
    QEvent,
    QMimeData,
    Qt,
    pyqtSignal,
)
from PyQt5.QtGui import (
    QDrag,
    QKeySequence,
)
from PyQt5.QtWidgets import (
    QApplication,
    QGridLayout,
    QLabel,
    QListWidget,
    QWidget,
)

from meety import io
from meety.gui.info_dialog.files import TabFiles
from meety.gui.main_window.meeting_list.context import MeetingItemMenu
from meety.gui.main_window.meeting_list.delegate import ItemDelegate
from meety.gui.main_window.meeting_list.item import MeetingItem
from meety.gui.main_window.options import OptionsWidget


class Meetings(QWidget):
    meeting_chosen = pyqtSignal(object, name="meeting_chosen")
    handler_chosen = pyqtSignal(object, name="handler_chosen")
    reload_requested = pyqtSignal(name="reload_requested")
    rating_mode_toggled = pyqtSignal(bool, name="rating_mode_toggled")

    def __init__(self):
        super().__init__()
        self.name = "Meetings"
        self._init_style()
        self._add_widgets()
        self._connect_widget_signals()
        self.update_rated_meetings([])

    def _init_style(self):
        pass

    def _add_widgets(self):
        layout = self._create_layout()
        layout.addWidget(self._create_info(), 0, 0)
        layout.addWidget(self._create_options(), 0, 1)
        layout.addWidget(self._create_list(), 1, 0, 1, 2)
        layout.addWidget(self._create_file_status(), 2, 0, 1, 2)
        self.setLayout(layout)

    def _create_layout(self):
        layout = QGridLayout()
        layout.setRowStretch(0, 0)
        layout.setRowStretch(1, 1)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        return layout

    def _create_info(self):
        self._info = QLabel("Meetings")
        self._info.setStyleSheet("""
            font: bold;
            margin-top: 5px;
        """)
        return self._info

    def _create_options(self):
        self._options = OptionsWidget()
        return self._options

    def _create_list(self):
        self._list = MeetingList()
        return self._list

    def _create_file_status(self):
        self._file_status = TabFiles()
        return self._file_status

    def _connect_widget_signals(self):
        self._list.itemActivated.connect(self._choose_meeting)
        self._options.rating_mode_toggled.connect(
            self.rating_mode_toggled.emit
        )
        self._list.installEventFilter(self)
        delegate = self._list.get_item_delegate()
        delegate.context_menu_requested.connect(
            self._context_menu_from_delegate
        )

    @property
    def current_meeting(self):
        return self._list.current_meeting

    def set_name(self, name):
        self.name = name
        self._update_info()

    def update_rated_meetings(self, rated_meetings):
        self._rated_meetings = rated_meetings
        self._update_info()
        self._update_entries(rated_meetings)

    def _update_info(self):
        num = len(self._rated_meetings)
        self._info.setText(f"{self.name} ({num})")

    def _update_entries(self, rated_meetings):
        row = self._list.currentRow()
        last = len(rated_meetings) - 1
        self._list.clear()
        for rmeeting in rated_meetings:
            item = MeetingItem(rmeeting)
            self._list.addItem(item)
        self._list.setCurrentRow(min(row, last))
        if rated_meetings:
            self._file_status.hide()

    def _choose_meeting(self, item):
        rmeeting = item.data(Qt.UserRole)
        self.meeting_chosen.emit(rmeeting.meeting)

    def eventFilter(self, source, event):
        if event.type() == QEvent.ContextMenu and source is self._list:
            return self._context_menu_on_list(source, event)
        return super().eventFilter(source, event)

    def _context_menu_from_delegate(self, event):
        return self._context_menu_on_list(self._list, event)

    def _context_menu_on_list(self, source, event):
        item = source.itemAt(event.pos())
        meeting = None
        if item:
            meeting = item.data(Qt.UserRole).meeting
        menu = MeetingItemMenu(meeting)
        menu.handler_chosen.connect(self.handler_chosen.emit)
        menu.reload_requested.connect(self.reload_requested.emit)
        menu.exec_(event.globalPos())
        return True

    def move_selection_down(self):
        self._list.move_selection_down()

    def move_selection_up(self):
        self._list.move_selection_up()


class MeetingList(QListWidget):
    COPY_COMPLETE_TIMESPAN = 2000

    def __init__(self):
        super().__init__()
        self.setProperty("class", "meetings")
        self._set_delegate()
        self._set_drag_and_drop()
        self._copy_tracker = EventTimingTracker(self.COPY_COMPLETE_TIMESPAN)
        self.setCurrentRow(0)

    def get_item_delegate(self):
        return self._item_delegate

    def _set_delegate(self):
        self._item_delegate = ItemDelegate()
        super().setItemDelegate(self._item_delegate)
        self._item_delegate.setParent(self)

    def move_selection_down(self):
        self.move_selection_by(+1)

    def move_selection_up(self):
        self.move_selection_by(-1)

    def move_selection_by(self, offset):
        self.move_selection_to(self.currentRow() + offset)

    def move_selection_to(self, row):
        new_row = self._valid_selection(row)
        if new_row is not None:
            super().setCurrentRow(new_row)

    def _valid_selection(self, row):
        start_selection = 0 if self.count() > 0 else None
        return io.utils.ensure_between(
            value=row,
            at_least=0,
            at_most=super().count()-1,
            on_wrong_type=start_selection
        )

    def _set_drag_and_drop(self):
        super().setDragEnabled(True)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.matches(QKeySequence.Copy):
            close_to_last_copy = self._copy_tracker.new_event()
            self._copy_to_clipboard(complete=close_to_last_copy)

    def _copy_to_clipboard(self, complete):
        mime_data = self._get_mime_data(complete)
        QApplication.clipboard().setMimeData(mime_data)

    def mouseMoveEvent(self, event):
        drag = QDrag(self)
        complete = event.modifiers() & Qt.ShiftModifier
        drag.setMimeData(self._get_mime_data(complete))
        drag.exec_(Qt.MoveAction)

    def _get_mime_data(self, complete):
        mime_data = QMimeData()
        if complete:
            self._set_mime_data_complete(mime_data)
        else:
            self._set_mime_data_minimal(mime_data)
        return mime_data

    def _set_mime_data_complete(self, mime_data):
        meeting = self.current_meeting
        mime_data.setText(meeting.input_data_to_str())

    def _set_mime_data_minimal(self, mime_data):
        meeting = self.current_meeting
        urls = meeting.urls
        if len(urls) == 0:
            url_text = "<NO URLs>"
        elif len(urls) == 1:
            url_text = urls[0]
        else:
            url_text = "\n" + "\n".join([f"- {u}" for u in urls])
        mime_data.setText(f"{meeting.name}: {url_text}")

    @property
    def current_meeting(self):
        return self.currentItem().data(Qt.UserRole).meeting


class EventTimingTracker:
    """Register new events and check whether the last event happened
    within the given time span."""

    def __init__(self, time_span):
        self._last = 0
        self._time_span = time_span

    """Returns True if and only if the new event is close to the last."""

    def new_event(self):
        current = time.time() * 1000
        close = current - self._last < self._time_span
        self._last = current
        return close
